(* $Id: godi_plist.ml 651 2006-10-08 19:55:56Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2004 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_versions
open Printf

type core_plist_element =
    [ `Dirrm of string
    | `File of string * string option   (* file name, checksum *)
    | `Comment of string
    | `Exec of string
    | `Unexec of string
    | `Option of [ `Extract_in_place | `Preserve ]
    | `Ignore
    | `Unsupported of string * string
    ]

type info_plist_element =
    [ `Name of string
    | `Cwd of string
    | `Src of string
    | `Pkgdep of Godi_versions.dependency
    | `Blddep of Godi_versions.dependency
    | `Pkgcfl of Godi_versions.dependency
    | `Display of string
    ]

type inst_plist_element =
    [ info_plist_element
    | core_plist_element
    ]

type user_plist_element =
    [ `Findlib    of user_plist_option * string
    | `Dllfindlib of user_plist_option * string
    | `Flatdir    of user_plist_option * string
    | `Deepdir    of user_plist_option * string
    | `Glob       of user_plist_option * string
    | `Library    of user_plist_option * string * string
    | `Stubdll    of user_plist_option * string * string
    | `File       of user_plist_option * string  (* no checksum *)
    | `Comment    of string
    | `Exec       of string
    | `Unexec     of string
    | `Option     of [ `Extract_in_place | `Preserve ]
    | `Dirrm      of string
    | `Unsupported of string * string
    ]

and user_plist_option =
    { upo_optional : bool;
      upo_ocamlopt : upo_bool;
      upo_ocamloptp : upo_bool;
      upo_ocamldll : upo_bool;
    }

and upo_bool = [ `None | `IfAvail | `IfNotAvail ]
 
type parsed_plist_element =
    (* Returned by parse_plist *)
    [ user_plist_element
    | info_plist_element
    | `Ignore
    ]

exception Plist_error of string * string
exception Plist_processing_error of string * string

let word_re = Str.regexp "[\r\n\t ]+";;

let supported_elements =
  [ "@name"; "@cwd"; "@cd"; "@src"; "@pkgdep"; "@blddep"; "@pkgcfl"; "@display";
    "@dirrm"; "@comment"; "@exec"; "@unexec"; "@option"; "@ignore";
    "@findlib"; "@dllfindlib"; "@flatdir"; "@deepdir"; "@glob";
    "@library"; "@stubdll";
    "@optional"; "@ocamlopt"; "@ocamloptp"; "@ocamldll"
  ] ;;

let default_upo =
  { upo_optional = false;
    upo_ocamlopt = `None;
    upo_ocamloptp = `None;
    upo_ocamldll = `None
  } ;;

let parse_plist ?(no_files = false) mode file next_line =
  (* Parses the file, but `File elements do not include checksums.
   * The checksum are represented as in the file, i.e. as comments.
   *
   * mode=`Inst: Allow only directives used by inst_plist_element
   * mode=`User: Allow only directives used by user_plist_element
   *)
  let rec parse() =
    try
      let line = next_line () in  (* or End_of_file *)
      let butfirst_word() =
	match Str.bounded_split word_re line 2 with
	    [] | [ _ ] -> ""
	  | [ _; word ] -> word
	  | _ -> assert false
      in
      let parse_dep s =
	try parse_dependency s
	with Failure _ ->
	  raise(Plist_error(file, line))
      in
      let rec parse_words upo words =
	let no_upo() =
	  if upo <> default_upo then
	    raise(Plist_error(file, line)) in
	let no_inst() =
	  if mode = `Inst then raise(Plist_error(file, line)) in
	let no_user() =
	  if mode = `User then raise(Plist_error(file, line))
	in
	if ((words <> [] && (List.hd words).[0] = '@') &&
	      not (List.mem (List.hd words) supported_elements)) then
	  `Unsupported(List.hd words, butfirst_word()) :: parse ()
	else
	  match words with
	    | [ "@name"; name ] -> 
		no_user(); no_upo(); `Name name :: parse ()
	    | [ ( "@cwd" | "@cd"); dir ] -> 
		no_user(); no_upo(); `Cwd dir :: parse ()
	    | [ "@src"; dir ] -> 
		no_user(); no_upo(); `Src dir :: parse ()
	    | [ "@pkgdep"; s ] -> 
		no_user(); no_upo(); `Pkgdep(parse_dep s) :: parse ()
	    | [ "@blddep"; s ] -> 
		no_user(); no_upo(); `Blddep(parse_dep s) :: parse ()
	    | [ "@pkgcfl"; s ] -> 
		no_user(); no_upo(); `Pkgcfl(parse_dep s) :: parse ()
	    | [ "@dirrm"; dir ] ->
		no_upo();
		if no_files then parse () else `Dirrm dir :: parse ()
	    | "@comment" :: _ ->
		no_upo();
		if no_files then parse () else
		  let rest = butfirst_word() in `Comment rest :: parse ()
	    | "@exec" :: _ ->
		no_upo();
		let rest = butfirst_word() in `Exec rest :: parse ()
	    | "@unexec" :: _ ->
		no_upo();
		if no_files then parse () else
		  let rest = butfirst_word() in `Unexec rest :: parse ()
	    | [ "@display"; file ] ->
		no_user(); no_upo();
		if no_files then parse () else
		  `Display file :: parse ()
	    | [ "@option"; "extract-in-place" ] ->
		no_upo();
		`Option `Extract_in_place :: parse ()
	    | [ "@option"; "preserve" ] ->
		no_upo();
		`Option `Preserve :: parse ()
	    | [ "@option"; arg ] ->
		no_upo();
		(`Unsupported("@option", arg)) :: parse()
	    | [ "@ignore" ] ->
		no_user(); no_upo();
		`Ignore :: parse()
	    | [ "@findlib"; name ] ->
		no_inst(); 
		if no_files then parse() else (`Findlib(upo, name)) :: parse()
	    | [ "@dllfindlib"; name ] ->
		no_inst(); 
		if no_files then parse() else (`Dllfindlib(upo, name)) :: parse()
	    | [ "@flatdir"; name ] ->
		no_inst(); 
		if no_files then parse() else (`Flatdir(upo, name)) :: parse()
	    | [ "@deepdir"; name ] ->
		no_inst(); 
		if no_files then parse() else (`Deepdir(upo, name)) :: parse()
	    | [ "@glob"; expr ] ->
		no_inst(); 
		if no_files then parse() else (`Glob(upo, expr)) :: parse()
	    | [ "@library"; dir; name ] ->
		no_inst(); 
		if no_files then parse() else (`Library(upo, dir, name)) :: parse()
	    | [ "@stubdll"; dir; name ] ->
		no_inst(); 
		if no_files then parse() else (`Stubdll(upo, dir, name)) :: parse()
	    | "@optional" :: rest ->
		no_inst(); 
		parse_words { upo with upo_optional = true } rest
	    | "@ocamlopt" :: rest ->
		no_inst(); 
		parse_words { upo with upo_ocamlopt = `IfAvail } rest
	    | "@!ocamlopt" :: rest ->
		no_inst(); 
		parse_words { upo with upo_ocamlopt = `IfNotAvail } rest
	    | "@ocamloptp" :: rest ->
		no_inst(); 
		parse_words { upo with upo_ocamloptp = `IfAvail } rest
	    | "@!ocamloptp" :: rest ->
		no_inst(); 
		parse_words { upo with upo_ocamloptp = `IfNotAvail } rest
	    | "@ocamldll" :: rest ->
		no_inst(); 
		parse_words { upo with upo_ocamldll = `IfAvail } rest
	    | "@!ocamldll" :: rest ->
		no_inst(); 
		parse_words { upo with upo_ocamldll = `IfNotAvail } rest
	    | [ file ] when file.[0] <> '@' ->
		if no_files then parse () else
		  `File(upo, file) :: parse ()
	    | [] -> parse ()
	    | _ -> raise(Plist_error(file, line))
		(* Wrong number of arguments! *)
      in
      let words = Str.split word_re line in
      parse_words default_upo words
    with
	End_of_file -> []
  in
  parse()
;;


let from_file parse_fun file =
  let f = open_in file in
  let next_line() = input_line f in
  try
    let r = parse_fun next_line in
    close_in f;
    r
  with error ->
    close_in f;
    raise error
;;


let from_string parse_fun s =
  let lines = ref (Str.split_delim (Str.regexp "\n") s) in
  let next_line() =
    match !lines with
	[] -> raise End_of_file
      | line :: lines' ->
	  lines := lines';
	  line
  in
  parse_fun next_line
;;


let rec apply_ignore_directive pl =
  match pl with
      `Ignore :: _ :: rest ->
	apply_ignore_directive rest
    | directive :: rest ->
	directive :: apply_ignore_directive rest
    | [] ->
	[]
;;


let path_re = Str.regexp "/+";;


let rec apply_rel_path path_l rel_path_l =
  match rel_path_l with
    | "." :: rel_path_l' ->
	apply_rel_path path_l rel_path_l'
    | ".." :: rel_path_l' ->
	if path_l = [] then
	  raise(Plist_processing_error("<unknown>",
				       "@cwd: Moving out of managed directory"));
	apply_rel_path (List.rev(List.tl(List.rev path_l))) rel_path_l'
    | d :: rel_path_l' ->
	apply_rel_path (path_l @ [d]) rel_path_l'
    | [] ->
	path_l
;;


let ph_pat_re = Str.regexp "%[D]" ;;

let rewrite_placeholders rel_path pattern  =
  if rel_path = "" then
    pattern
  else
    Str.global_substitute
      ph_pat_re
      (fun s ->
	 let m = Str.matched_string s in
	 match m with
	   | "%D" -> "%D/" ^ rel_path
	   | _    -> assert false
      )
      pattern
;;


let apply_cwd_directive pl =
  let rec apply have_cwd abs_prefix rel_prefix rel_prefix_l pl =
    match pl with
      | (`Cwd path) as e :: rest ->
	  if have_cwd then (
	    if path.[0] = '/' then (
	      if path <> abs_prefix then
		raise(Plist_processing_error
			("<unknown>",
			 "@cwd must have relative path (except the first one)"))
	      else 
		(* path = abs_prefix allowed for backwards-compatibility *)
		apply have_cwd abs_prefix "" [] rest
	    )
	    else
	      let l = Str.split path_re path in
	      let rel_prefix_l' = apply_rel_path rel_prefix_l l in
	      let rel_prefix' = String.concat "/" rel_prefix_l' in
	      apply have_cwd abs_prefix rel_prefix' rel_prefix_l' rest
	  )
	  else
	    (* First @cwd *)
	    e :: apply true path "" [] rest
      | (`File(name,chksum)) :: rest ->
	  (`File(Filename.concat rel_prefix name, chksum)) :: 
	    apply have_cwd abs_prefix rel_prefix rel_prefix_l rest
      | (`Dirrm name) :: rest ->
	  (`Dirrm(Filename.concat rel_prefix name)) ::
	    apply have_cwd abs_prefix rel_prefix rel_prefix_l rest
      | (`Exec cmd) :: rest ->
	  let cmd' = rewrite_placeholders rel_prefix cmd in
	  (`Exec cmd') 
	  :: apply have_cwd abs_prefix rel_prefix rel_prefix_l rest
      | (`Unexec cmd) :: rest ->
	  let cmd' = rewrite_placeholders rel_prefix cmd in
	  (`Unexec cmd') 
	  :: apply have_cwd abs_prefix rel_prefix rel_prefix_l rest
      | (#inst_plist_element as e) :: rest ->
	  e :: apply have_cwd abs_prefix rel_prefix rel_prefix_l rest
      | [] -> []
  in
  apply false "" "" [] pl
;;


let md5_syntax_re = Str.regexp "^MD5:[0-9a-f]+$" ;;

let parse_inst_plist_from ?no_files file next_line =
  (* Also parse MD5 checksums: *)

  let md5_syntax s =
    Str.string_match md5_syntax_re s 0 in

  let rec map_to_inst have_cwd elements =
    match elements with
      | (`Cwd path) as e :: rest ->
	  let abs = path <> "" && path.[0] = '/' in
	  if not have_cwd && not abs then
	    raise(Plist_processing_error
		    (file, "First @cwd must have absolute path"))
(* There are currently old packages with several identical @cwd <abs_path>
 * around, so disable this check for now:
 *)
(*	      
	  else
	    if have_cwd && abs then
	      raise(Plist_processing_error
		      (file, 
		       "@cwd must have relative path (except the first one)"))
 *)
	  ;
	  e :: map_to_inst true rest
      | (#info_plist_element as e) :: rest ->
	  e :: map_to_inst have_cwd rest
      | `Findlib(_,_) :: _
      | `Dllfindlib(_,_) :: _
      | `Flatdir(_,_) :: _
      | `Deepdir(_,_) :: _
      | `Glob(_,_) :: _
      | `Library(_,_,_) :: _
      | `Stubdll(_,_,_) :: _ ->
	  assert false  (* Not possible with mode=`Inst *)
      | `Comment s :: rest ->
	  (`Comment s) :: map_to_inst have_cwd rest
      | `Exec s :: rest ->
	  if not have_cwd then
	    raise(Plist_processing_error
		    (file, "There must be @cwd before first @exec"));
	  (`Exec s) :: map_to_inst have_cwd rest
      | `Unexec s :: rest ->
	  if not have_cwd then
	    raise(Plist_processing_error
		    (file, "There must be @cwd before first @unexec"));
	  (`Unexec s) :: map_to_inst have_cwd rest
      | `Option p :: rest ->
	  (`Option p) :: map_to_inst have_cwd rest
      | `Dirrm s :: rest ->
	  if not have_cwd then
	    raise(Plist_processing_error
		    (file, "There must be @cwd before first @dirrm"));
	  (`Dirrm s) :: map_to_inst have_cwd rest
      | `Unsupported(opt,arg) :: rest ->
	  (`Unsupported(opt,arg)) :: map_to_inst have_cwd rest
      | `File(_,file) :: `Comment c :: rest when md5_syntax c ->
	  if not have_cwd then
	    raise(Plist_processing_error
		    (file, "There must be @cwd before first file"));
	  `File(file, Some c) :: map_to_inst have_cwd rest
      | `File(_,file) :: rest ->
	  if not have_cwd then
	    raise(Plist_processing_error
		    (file, "There must be @cwd before first file"));
	  `File(file, None) :: map_to_inst have_cwd rest
      | `Ignore :: rest ->
	  `Ignore :: map_to_inst have_cwd rest
      | [] ->
	  []
  in
  apply_cwd_directive
    (apply_ignore_directive
       (map_to_inst 
	  false
	  (parse_plist ?no_files `Inst file next_line)))
;;


let parse_inst_plist ?no_files file =
  from_file (parse_inst_plist_from ?no_files file) file ;;


let parse_inst_plist_string ?no_files str =
  from_string (parse_inst_plist_from ?no_files "<nofile>") str ;;


let parse_user_plist_from  ?no_files file next_line =
  let rec map_to_user elements =
    match elements with
      | (#user_plist_element as e) :: rest ->
	  e :: map_to_user rest
      | _ :: _ ->
	  assert false  (* Not possible with mode=`User *)
      | [] ->
	  []
  in
  map_to_user (parse_plist ?no_files `User file next_line)
;;


let parse_user_plist file =
  from_file (parse_user_plist_from file) file ;;


let extract_package_name pl =
  match 
    List.find (function `Name n -> true | _ -> false) pl
  with
      `Name n -> n
    | _ -> assert false
;;


let extract_prefix pl =
  match
    List.find (function `Cwd p -> true | _ -> false) pl
  with
    | `Cwd p -> p
    | _ -> assert false
;;


let extract_display pl =
  match
    List.find (function `Display p -> true | _ -> false) pl
  with
    | `Display d -> d
    | _ -> assert false
;;


let extract_pkgdep pl =
  List.map
    (function `Pkgdep dep -> dep | _ -> assert false)
    (List.filter
       (function `Pkgdep _ -> true | _ -> false)
       pl)
;;


let extract_blddep pl =
  List.map
    (function `Blddep dep -> dep | _ -> assert false)
    (List.filter
       (function `Blddep _ -> true | _ -> false)
       pl)
;;


let extract_pkgcfl pl =
  List.map
    (function `Pkgcfl dep -> dep | _ -> assert false)
    (List.filter
       (function `Pkgcfl _ -> true | _ -> false)
       pl)
;;


let print_inst_plist file plist =
  let f = open_out file in
  try
    List.iter
      (function
	   `Name name ->
	     fprintf f "@name %s\n" name
	 | `Cwd dir ->
	     fprintf f "@cwd %s\n" dir
	 | `Src dir ->
	     fprintf f "@src %s\n" dir
	 | `Pkgdep d ->
	     fprintf f "@pkgdep %s\n" (print_dependency d)
	 | `Blddep d ->
	     fprintf f "@blddep %s\n" (print_dependency d)
	 | `Pkgcfl d ->
	     fprintf f "@pkgcfl %s\n" (print_dependency d)
	 | `Dirrm dir ->
	     fprintf f "@dirrm %s\n" dir
	 | `File(name,None) ->
	     fprintf f "%s\n" name
	 | `File(name,Some cs) ->
	     fprintf f "%s\n@comment %s\n" name cs
	 | `Comment s ->
	     fprintf f "@comment %s\n" s
	 | `Exec s ->
	     fprintf f "@exec %s\n" s
	 | `Unexec s ->
	     fprintf f "@unexec %s\n" s
	 | `Display file ->
	     fprintf f "@display %s\n" file
	 | `Option `Extract_in_place ->
	     fprintf f "@option extract-in-place\n"
	 | `Option `Preserve ->
	     fprintf f "@option preserve\n"
	 | `Ignore ->
	     fprintf f "@ignore\n"
	 | `Unsupported(directive,"") ->
	     fprintf f "%s\n" directive
	 | `Unsupported(directive,arg) ->
	     fprintf f "%s %s\n" directive arg
      )
      plist;
    close_out f
  with
      error ->
	close_out f; raise error
;;


let prepare_for_pkgdb plist =
  List.filter
    (function
       | `Pkgcfl _ -> false
       | _ -> true
    )
    plist


exception Plist_expansion_error of string


let expand_plist ~root ~ocamlopt_feature ~ocamloptp_feature ~ocamldll_feature
                 ~library_re ~stubdll_re ~cygwin_fixups
                 elements =

  let if_upo upo file_opt f_then f_else =
    (* Call f only if upo allows it *)
    let if_feature cond feat =
      match cond with
	| `None -> true
	| `IfAvail -> feat
	| `IfNotAvail -> not feat in
    let eff =
      if_feature upo.upo_ocamlopt ocamlopt_feature &&
	if_feature upo.upo_ocamloptp ocamloptp_feature &&
	if_feature upo.upo_ocamldll ocamldll_feature &&
	( match file_opt with
	    | None -> true
	    | Some file -> 
		not upo.upo_optional || Sys.file_exists file
	) in
    if eff then f_then() else f_else()
  in

  let output_files ?(ignore_directories=false) files : core_plist_element list =
    (* - [files] is the path relative to [root]
     * - Filters out directories if requested, else fail on directories
     * - Cygwin: Performs the .exe magic (if <file> and <file>.exe both
     *   exist and point to the same file, take <file>.exe as name)
     * - computes checksums for regular files
     * - returns list of [`File] directives
     *)
    let files' =
      (* Cygwin fixups part 1:
       * If there both <file> and <file>.exe in the list, simply remove <file>.
       * For example, this fixes "@glob file{,.exe}".
       *)
      if cygwin_fixups then
	List.filter
	  (fun file ->
	     not (List.mem (file ^ ".exe") files))
	  files
      else
	files in
    let file_dirs =
      List.map
	(fun file ->
	   let path = Filename.concat root file in
	   if Godi_file.file_type ~is_symlink:true path then (
	     `File(file, None)
	   )
	   else
	     if Godi_file.file_type ~is_dir:true path then (
	       if ignore_directories then
		 `File("",None)   (* Filtered out by List.filter above *)
	       else
		 raise(Plist_expansion_error("Directory found: " ^ path))
	     )
	     else
	       if Godi_file.file_type ~is_regular:true path then (
		 let sum = "MD5:" ^ Digest.to_hex (Digest.file path) in
		 `File(file, Some sum)
	       )
	       else
		 if Godi_file.file_exists path then
		   raise(Plist_expansion_error("Unknown file type found: " ^ path))
		 else
		   raise(Plist_expansion_error("File not found: " ^ path))
	)
	files' in
    let file_dirs' =
      if cygwin_fixups then
	(* Cygwin fixups part 2:
         * If there is a <file> in the list, and <file> and <file>.exe
         *    exist and are the same file, change <file> to <file>.exe.
         *)
	List.map
	  (function 
	     | (`File(name, chksum)) ->
		 let name_exe = name ^ ".exe" in
		 let path = Filename.concat root name in
		 let path_exe = Filename.concat root name_exe in
		 let ok = 
		   name <> "" &&
		     Godi_file.file_type ~is_regular:true ~is_not_symlink:true path in
		 let ok_exe = 
		   Godi_file.file_type ~is_regular:true ~is_not_symlink:true 
		     path_exe in
		 if ok && ok_exe && Godi_file.same_file path path_exe then
		   `File(name_exe, chksum)
		 else
		   `File(name, chksum)
	     | other -> other   (* for typing reasons *)
	  )
	  file_dirs
      else
	file_dirs in

    List.filter
      (function 
	 | (`File(name, _)) -> name <> ""
	 | _ -> true (* for typing reasons *)
      )
      file_dirs'
  in

  let expand_findlib upo name f_add =
    let findlib_path = Filename.concat "lib/ocaml/pkg-lib" name in
    let dir = Filename.concat root findlib_path in
    if_upo upo (Some dir) 
      (fun () ->
	 if not (Sys.file_exists dir) then
	   raise(Plist_expansion_error("Findlib package is missing: " ^ 
					 name));
	 let files = Godi_file.read_dir dir in
	 output_files
	   (List.map
	      (fun file -> Filename.concat findlib_path file)
	      files)
	 @ f_add()
      )
      (fun () -> [] )
  in

  let expand_dllfindlib upo name =
    expand_findlib
      upo name
      (fun () ->
	 let dllfindlib_path = "lib/ocaml/pkg-lib/stublibs" in
	 let dllfindlib_full_path = Filename.concat root dllfindlib_path in
	 let files = Godi_file.read_dir dllfindlib_full_path in
	 let owner_files =
	   List.filter
	     (fun file -> 
		let lines = 
		  Godi_file.read_file_lines 
		    (Filename.concat dllfindlib_full_path file) in
		Filename.check_suffix file ".owner"
		  && lines <> [] 
		  && List.hd lines = name
	     )
	     files in
	 List.iter
	   (fun owner_file ->
	      let base_file = Filename.chop_suffix owner_file ".owner" in
	      let f =
		Filename.concat dllfindlib_full_path base_file in
	      if base_file = "" || not (Sys.file_exists f) then
		raise(Plist_expansion_error("File missing: " ^ f));
	   )
	   owner_files;
	 let files' =
	   List.flatten
	     ( List.map
		 (fun owner_file -> 
		    [ Filename.concat dllfindlib_path owner_file;
		      Filename.concat
			dllfindlib_path
			(Filename.chop_suffix owner_file ".owner") 
		    ]
		 )
		 owner_files) in
	 output_files files' 
      )
  in

  let expand_flatdir upo path =
    let dir = Filename.concat root path in
    if_upo upo (Some dir) 
      (fun () ->
	 if not (Sys.file_exists dir) then
	   raise(Plist_expansion_error("Directory is missing: " ^ dir));
	 let files = Godi_file.read_dir dir in
	 output_files
	   ~ignore_directories:true
	   (List.map
	      (fun file -> Filename.concat path file)
	      files)
      )
      (fun () -> [] )
  in

  let expand_deepdir upo path =
    let dir = Filename.concat root path in
    if_upo upo (Some dir) 
      (fun () ->
	 if not (Sys.file_exists dir) then
	   raise(Plist_expansion_error("Directory is missing: " ^ dir));
	 let files = Godi_file.read_dir_hierarchy dir in
	 output_files
	   ~ignore_directories:true
	   (List.map
	      (fun file -> Filename.concat path file)
	      files)
      )
      (fun () -> [] )
  in

  let expand_glob upo expr =
    let gexpr = Godi_match.parse_glob_expr expr in
    if_upo upo None
      (fun () ->
	 (* @glob is always "@optional" *)
	 let files = 
	   Godi_match.glob 
	     ~base_dir:root
	     ~mode:`Existing_paths
	     gexpr in
	 output_files ~ignore_directories:true files
      )
      (fun () -> [] )
  in

  let expand_lib lib_re upo dir name =
    let full_dir = Filename.concat root dir in
    let re = lib_re name in
    if_upo upo (Some full_dir) 
      (fun () ->
	 if not (Sys.file_exists dir) then
	   raise(Plist_expansion_error("Directory is missing: " ^ full_dir));
	 let files = Godi_file.read_dir dir in
	 let files' =
	   List.filter
	     (fun file -> Str.string_match re file 0)
	     files in
	 let out =
	   output_files
	     ~ignore_directories:true
	     (List.map
		(fun file -> Filename.concat dir file)
		files') in
	 if out = [] && not upo.upo_optional then
	   raise(Plist_expansion_error("No library <" ^ name ^ 
					 "> found in directory: " ^ full_dir));
	 out
      )
      (fun () -> [] )
  in

  let no_abs_path p =
    if p <> "" && p.[0] = '/' then
      raise(Plist_expansion_error("Absolute paths not allowed: " ^ p))
  in

  let rec expand elements : core_plist_element list =
    match elements with
      | `Findlib(upo,name) :: rest ->
	  expand_findlib upo name (fun () -> []) @ expand rest
      | `Dllfindlib(upo,name) :: rest ->
	  expand_dllfindlib upo name @ expand rest
      | `Flatdir(upo,name) :: rest ->
	  no_abs_path name;
	  expand_flatdir upo name @ expand rest
      | `Deepdir(upo,name) :: rest ->
	  no_abs_path name;
	  expand_deepdir upo name @ expand rest
      | `Glob(upo,expr) :: rest ->
	  no_abs_path expr;
	  expand_glob upo expr @ expand rest
      | `Library(upo,dir,name) :: rest ->
	  no_abs_path dir;
	  expand_lib library_re upo dir name @ expand rest
      | `Stubdll(upo,dir,name) :: rest ->
	  no_abs_path dir;
	  expand_lib stubdll_re upo dir name @ expand rest
      | `File(upo,name) :: rest ->
	  no_abs_path name;
	  output_files [ name ] @ expand rest
      | `Comment s :: rest ->
	  (`Comment s) :: expand rest
      | `Exec s :: rest ->
	  (`Exec s) :: expand rest
      | `Unexec s  :: rest->
	  (`Unexec s) :: expand rest
      | `Option p :: rest ->
	  (`Option p) :: expand rest
      | `Dirrm s :: rest ->
	  expand rest
      | `Unsupported (opt, arg) :: rest ->
	  raise(Plist_expansion_error("Unknown directive: " ^ opt))
      | [] -> 
	  []
  in
  expand elements
;;
